#!/bin/bash
#

# build_trtis_clients.sh

# if this build script complains that it cannot find bzlib.h,
# you need to install bzip2-devel on Scientific Linux

usage()
{
   echo "USAGE: `basename ${0}` <product_directory> <e19|c7> <debug|prof> [tar]"
}

# -------------------------------------------------------------------
# shared boilerplate
# -------------------------------------------------------------------

get_this_dir()
{
    ( cd / ; /bin/pwd -P ) >/dev/null 2>&1
    if (( $? == 0 )); then
      pwd_P_arg="-P"
    fi
    reldir=`dirname ${0}`
    thisdir=`cd ${reldir} && /bin/pwd ${pwd_P_arg}`
}

get_ssibuildshims()
{
    # make sure we can use the setup alias
    if [ -z ${UPS_DIR} ]
    then
       echo "ERROR: please setup ups"
       exit 1
    fi
    source `${UPS_DIR}/bin/ups setup ${SETUP_UPS}`

    PRODUCTS=${product_dir}:${PRODUCTS}

    setup ssibuildshims ${ssibuildshims_version}
    if [ -z ${SSIBUILDSHIMS_DIR} ]
    then
       echo "ERROR: failed to setup ssibuildshims ${ssibuildshims_version}"
       exit 1
    fi
}

# -------------------------------------------------------------------
# start processing
# -------------------------------------------------------------------

product_dir=${1}
basequal=${2}
extraqual=${3}
maketar=${4}

case ${extraqual} in
  debug) buildtype=Debug;;
  prof)  buildtype=MinSizeRel;;
  opt)   buildtype=Release;;
  * )
    echo "ERROR: please specify debug, opt, or prof"
    usage
    exit 1
    ;;
esac

if [ -z ${product_dir} ]
then
   echo "ERROR: please specify the local product directory"
   usage
   exit 1
fi

# -------------------------------------------------------------------
# package name and version
# -------------------------------------------------------------------

package=trtis_clients
origpkgver=v19_11
pkgver=${origpkgver}a
pkgdotver=`echo ${origpkgver} | sed -e 's/_/./g' | sed -e 's/^v/r/'`
ssibuildshims_version=v1_04_14
tensorrtdir=tensorrt-inference-server

get_this_dir

get_ssibuildshims

source define_basics --

if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} ${fullqual}
   exit 0
fi

echo "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${flvr})"

# -------------------------------------------------------------------
# package dependent stuff goes here
# -------------------------------------------------------------------

gcc_compilers="e19"
clang_compilers="c7"
std=17
cxxflg="-Wno-deprecated-declarations"
if [[ "${basequal}" =~ $gcc_compilers ]]
then
  cc=gcc
  cxx=g++
elif [[ "${basequal}" =~ $clang_compilers ]]
then
  cc=clang
  cxx=clang++
else
  ssi_die "Qualifier $basequal not recognized."
fi

# fake ups declare
fakedb=${product_dir}/${package}/${pkgver}/fakedb
${SSIBUILDSHIMS_DIR}/bin/fake_declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

setup -B ${package} ${pkgver} -q ${fullqual} -z ${fakedb}:${PRODUCTS} || ssi_die "fake setup failed"

# We need cmake
source ${SSIBUILDSHIMS_DIR}/bin/setup_cmake || ssi_die "ERROR: failed to setup camke"

if [ -z ${PROTOBUF_FQ_DIR} ]
then
   echo "ERROR: failed to setup protobuf"
   exit 1
else
   echo "protobuf is in ${PROTOBUF_FQ_DIR}"
fi

if [ -z ${OPENCV_FQ_DIR} ]
then
   echo "ERROR: failed to setup protobuf"
   exit 1
else
   echo "opencv is in ${OPENCV_FQ_DIR}"
fi

if [ -z ${PYTHON_DIR} ]
then
    echo "ERROR: failed to setup python"
    exit 1
else
    echo "python is in ${PYTHON_DIR}"
fi

echo "Building ${package} in ${pkgdir}"

mkdir -p ${pkgdir}
if [ ! -d ${pkgdir} ]
then
   echo "ERROR: failed to create ${pkgdir}"
   exit 1
fi

mkdir -p ${pkgdir}/include/trtis_clients
if [ ! -d ${pkgdir}/include/trtis_clients ]
then
   echo "ERROR: failed to create ${pkgdir}/include/trtis_clients"
   exit 1
fi

set -x

# Require several python facilities
python_shortdotversion=${PYTHON_VERSION#v}
python_shortdotversion=${python_shortdotversion//_/.}
python_shortdotversion=${python_shortdotversion%.*}

pip install --prefix $pkgdir wheel setuptools grpcio-tools
export PYTHONPATH=${pkgdir}/lib/python${python_shortdotversion}/site-packages

cd ${pkgdir}
tar -xf ${tardir}/${tensorrtdir}.tar || ssi_die "Failed to unwind ${tensorrtdir}.tar"

cd ${tensorrtdir}/build || ssi_die "cannot find build subdirectory"

# Apply patches
patch -b -p0 < ${patchdir}/use_ups.patch || ssi_die "Unable to apply ${patchdir}/use_ups.patch."
patch -b -p0 < ${patchdir}/fix_compile_flags.patch || ssi_die "Unable to apply ${patchdir}/fix_compile_flags.patch."

cmake -DCMAKE_BUILD_TYPE=${buildtype} \
      -DCMAKE_C_COMPILER=${cc} \
      -DCMAKE_CXX_COMPILER=${cxx} \
      -DCMAKE_CXX_FLAGS="${cxxflg}" \
      -DCMAKE_PREFIX_PATH=${OPENCV_FQ_DIR}/share/OpenCV \
      -DTRTIS_ENABLE_GPU=OFF
ncore=`${SSIBUILDSHIMS_DIR}/bin/ncores`
make -j ${ncore} trtis-clients  || ssi_die "make step failed"
set +x

# Manual install
cp -r trtis-clients/install/include/* ${pkgdir}/include/trtis_clients/
cp -r trtis-clients/install/lib/* ${pkgdir}/lib

# Clean up
cd ${pkgdir}
rm -rf ${tensorrtdir}

# real ups declare
${SSIBUILDSHIMS_DIR}/bin/declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

# -------------------------------------------------------------------
# common bottom stuff
# -------------------------------------------------------------------

# this should not complain
echo "Finished building ${package} ${pkgver}"
setup -B ${package} ${pkgver} -q ${fullqual} -z ${product_dir} || ssi_die "setup failed"
echo "${package} is installed at ${TRTIS_CLIENTS_FQ_DIR}"

# this must be last
if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} ${fullqual}
fi

exit 0
